@@ -1,9 +1,7 @@
-# coding=utf-8
-#
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis/
 #
-# Most of this work is copyright (C) 2013-2019 David R. MacIver
+# Most of this work is copyright (C) 2013-2020 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
 # CONTRIBUTING.rst for a full list of people who may hold copyright, and
 # consult the git log if you need to determine who owns an individual
@@ -15,58 +13,35 @@
 #
 # END HEADER
 
-from __future__ import absolute_import, division, print_function
-
 import math
 import re
-from collections import namedtuple
+from typing import Any, NamedTuple, Sequence, Tuple, Union
 
 import numpy as np
 
-import hypothesis._strategies as st
 import hypothesis.internal.conjecture.utils as cu
-from hypothesis import Verbosity, assume
-from hypothesis._settings import note_deprecation
+import hypothesis.strategies._internal.core as st
+from hypothesis import assume
 from hypothesis.errors import InvalidArgument
-from hypothesis.internal.compat import (
-    PY2,
-    hrange,
-    integer_types,
-    quiet_raise,
-    string_types,
-)
 from hypothesis.internal.coverage import check_function
-from hypothesis.internal.reflection import proxies, reserved_means_kwonly_star
+from hypothesis.internal.reflection import deprecated_posargs, proxies
 from hypothesis.internal.validation import check_type, check_valid_interval
-from hypothesis.reporting import current_verbosity
-from hypothesis.searchstrategy import SearchStrategy
+from hypothesis.strategies._internal import SearchStrategy
+from hypothesis.strategies._internal.strategies import T
 from hypothesis.utils.conventions import UniqueIdentifier, not_set
 
-if PY2:
-    BroadcastableShapes = namedtuple(
-        "BroadcastableShapes", ["input_shapes", "result_shape"]
-    )
-else:
-    from typing import NamedTuple, Tuple
-
-    Shape = Tuple[int, ...]  # noqa
-    BroadcastableShapes = NamedTuple(
-        "BroadcastableShapes",
-        [("input_shapes", Tuple[Shape, ...]), ("result_shape", Shape)],
-    )
-
-if False:
-    from typing import Any, Union, Sequence, Tuple  # noqa
-    from hypothesis.searchstrategy.strategies import T  # noqa
-
-    BasicIndex = Tuple[Union[int, slice, ellipsis, np.newaxis], ...]  # noqa
+Shape = Tuple[int, ...]
+BasicIndex = Tuple[Union[int, slice, "ellipsis", np.newaxis], ...]
+BroadcastableShapes = NamedTuple(
+    "BroadcastableShapes",
+    [("input_shapes", Tuple[Shape, ...]), ("result_shape", Shape)],
+)
 
 TIME_RESOLUTIONS = tuple("Y  M  D  h  m  s  ms  us  ns  ps  fs  as".split())
 
 
 @st.defines_strategy_with_reusable_values
-def from_dtype(dtype):
-    # type: (np.dtype) -> st.SearchStrategy[Any]
+def from_dtype(dtype: np.dtype) -> st.SearchStrategy[Any]:
     """Creates a strategy which can generate any value of the given dtype."""
     check_type(np.dtype, dtype, "dtype")
     # Compound datatypes, eg 'f4,f4,f4'
@@ -80,45 +55,48 @@ def from_dtype(dtype):
         return arrays(subtype, shape)
 
     # Scalar datatypes
-    if dtype.kind == u"b":
+    if dtype.kind == "b":
         result = st.booleans()  # type: SearchStrategy[Any]
-    elif dtype.kind == u"f":
+    elif dtype.kind == "f":
         if dtype.itemsize == 2:
             result = st.floats(width=16)
         elif dtype.itemsize == 4:
             result = st.floats(width=32)
         else:
             result = st.floats()
-    elif dtype.kind == u"c":
+    elif dtype.kind == "c":
         if dtype.itemsize == 8:
             float32 = st.floats(width=32)
             result = st.builds(complex, float32, float32)
         else:
             result = st.complex_numbers()
-    elif dtype.kind in (u"S", u"a"):
+    elif dtype.kind in ("S", "a"):
         # Numpy strings are null-terminated; only allow round-trippable values.
         # `itemsize == 0` means 'fixed length determined at array creation'
         result = st.binary(max_size=dtype.itemsize or None).filter(
             lambda b: b[-1:] != b"\0"
         )
-    elif dtype.kind == u"u":
+    elif dtype.kind == "u":
         result = st.integers(min_value=0, max_value=2 ** (8 * dtype.itemsize) - 1)
-    elif dtype.kind == u"i":
+    elif dtype.kind == "i":
         overflow = 2 ** (8 * dtype.itemsize - 1)
         result = st.integers(min_value=-overflow, max_value=overflow - 1)
-    elif dtype.kind == u"U":
+    elif dtype.kind == "U":
         # Encoded in UTF-32 (four bytes/codepoint) and null-terminated
         result = st.text(max_size=(dtype.itemsize or 0) // 4 or None).filter(
-            lambda b: b[-1:] != u"\0"
+            lambda b: b[-1:] != "\0"
         )
-    elif dtype.kind in (u"m", u"M"):
+    elif dtype.kind in ("m", "M"):
         if "[" in dtype.str:
             res = st.just(dtype.str.split("[")[-1][:-1])
         else:
+            # Note that this case isn't valid to pass to arrays(), but we support
+            # it here because we'd have to guard against equivalents in arrays()
+            # regardless and drawing scalars is a valid use-case.
             res = st.sampled_from(TIME_RESOLUTIONS)
         result = st.builds(dtype.type, st.integers(-(2 ** 63), 2 ** 63 - 1), res)
     else:
-        raise InvalidArgument(u"No strategy inference for {}".format(dtype))
+        raise InvalidArgument("No strategy inference for {}".format(dtype))
     return result.map(dtype.type)
 
 
@@ -132,14 +110,14 @@ def check_argument(condition, fail_message, *f_args, **f_kwargs):
 def order_check(name, floor, small, large):
     check_argument(
         floor <= small,
-        u"min_{name} must be at least {} but was {}",
+        "min_{name} must be at least {} but was {}",
         floor,
         small,
         name=name,
     )
     check_argument(
         small <= large,
-        u"min_{name}={} is larger than max_{name}={}",
+        "min_{name}={} is larger than max_{name}={}",
         small,
         large,
         name=name,
@@ -154,24 +132,26 @@ class ArrayStrategy(SearchStrategy):
         self.dtype = dtype
         self.element_strategy = element_strategy
         self.unique = unique
+        self._check_elements = dtype.kind not in ("O", "V")
 
     def set_element(self, data, result, idx, strategy=None):
         strategy = strategy or self.element_strategy
         val = data.draw(strategy)
-        result[idx] = val
-        if self._report_overflow and val != result[idx] and val == val:
-            note_deprecation(
+        try:
+            result[idx] = val
+        except TypeError as err:
+            raise InvalidArgument(
+                "Could not add element=%r of %r to array of %r - possible mismatch "
+                "of time units in dtypes?" % (val, val.dtype, result.dtype)
+            ) from err
+        if self._check_elements and val != result[idx] and val == val:
+            raise InvalidArgument(
                 "Generated array element %r from %r cannot be represented as "
                 "dtype %r - instead it becomes %r (type %r).  Consider using a more "
                 "precise strategy, for example passing the `width` argument to "
-                "`floats()`, as this will be an error in a future version."
-                % (val, strategy, self.dtype, result[idx], type(result[idx])),
-                since="2019-07-28",
+                "`floats()`."
+                % (val, strategy, self.dtype, result[idx], type(result[idx]))
             )
-            # Because the message includes the value of the generated element,
-            # it would be easy to spam users with thousands of warnings.
-            # We therefore only warn once per draw, unless in verbose mode.
-            self._report_overflow = current_verbosity() >= Verbosity.verbose
 
     def do_draw(self, data):
         if 0 in self.shape:
@@ -180,13 +160,9 @@ class ArrayStrategy(SearchStrategy):
         # Because Numpy allocates memory for strings at array creation, if we have
         # an unsized string dtype we'll fill an object array and then cast it back.
         unsized_string_dtype = (
-            self.dtype.kind in (u"S", u"a", u"U") and self.dtype.itemsize == 0
+            self.dtype.kind in ("S", "a", "U") and self.dtype.itemsize == 0
         )
 
-        # Reset this flag for each test case to emit warnings from set_element
-        # Skip the check for object or void (multi-element) dtypes
-        self._report_overflow = self.dtype.kind not in "OV" and not unsized_string_dtype
-
         # This could legitimately be a np.empty, but the performance gains for
         # that would be so marginal that there's really not much point risking
         # undefined behaviour shenanigans.
@@ -221,7 +197,7 @@ class ArrayStrategy(SearchStrategy):
                     else:
                         elements.reject()
             else:
-                for i in hrange(len(result)):
+                for i in range(len(result)):
                     self.set_element(data, result, i)
         else:
             # We draw numpy arrays as "sparse with an offset". We draw a
@@ -294,12 +270,11 @@ class ArrayStrategy(SearchStrategy):
             out = result.astype(self.dtype)
             mismatch = out != result
             if mismatch.any():
-                note_deprecation(
+                raise InvalidArgument(
                     "Array elements %r cannot be represented as dtype %r - instead "
                     "they becomes %r.  Use a more precise strategy, e.g. without "
                     "trailing null bytes, as this will be an error future versions."
-                    % (result[mismatch], self.dtype, out[mismatch]),
-                    since="2019-07-28",
+                    % (result[mismatch], self.dtype, out[mismatch])
                 )
             result = out
 
@@ -319,14 +294,15 @@ def fill_for(elements, unique, fill, name=""):
 
 
 @st.defines_strategy
+@deprecated_posargs
 def arrays(
-    dtype,  # type: Any
-    shape,  # type: Union[int, Shape, st.SearchStrategy[Shape]]
-    elements=None,  # type: st.SearchStrategy[Any]
-    fill=None,  # type: st.SearchStrategy[Any]
-    unique=False,  # type: bool
-):
-    # type: (...) -> st.SearchStrategy[np.ndarray]
+    dtype: Any,
+    shape: Union[int, Shape, st.SearchStrategy[Shape]],
+    *,
+    elements: st.SearchStrategy[Any] = None,
+    fill: st.SearchStrategy[Any] = None,
+    unique: bool = False
+) -> st.SearchStrategy[np.ndarray]:
     r"""Returns a strategy for generating :class:`numpy:numpy.ndarray`\ s.
 
     * ``dtype`` may be any valid input to :class:`~numpy:numpy.dtype`
@@ -412,12 +388,23 @@ def arrays(
     # From here on, we're only dealing with values and it's relatively simple.
     dtype = np.dtype(dtype)
     if elements is None:
+        if dtype.kind in ("m", "M") and "[" not in dtype.str:
+            # For datetime and timedelta dtypes, we have a tricky situation -
+            # because they *may or may not* specify a unit as part of the dtype.
+            # If not, we flatmap over the various resolutions so that array
+            # elements have consistent units but units may vary between arrays.
+            return (
+                st.sampled_from(TIME_RESOLUTIONS)
+                .map((dtype.str + "[{}]").format)
+                .flatmap(lambda d: arrays(d, shape=shape, fill=fill, unique=unique))
+            )
         elements = from_dtype(dtype)
-    if isinstance(shape, integer_types):
+    check_type(SearchStrategy, elements, "elements")
+    if isinstance(shape, int):
         shape = (shape,)
     shape = tuple(shape)
     check_argument(
-        all(isinstance(s, integer_types) for s in shape),
+        all(isinstance(s, int) for s in shape),
         "Array shape must be integer in each dimension, provided shape was {}",
         shape,
     )
@@ -426,11 +413,13 @@ def arrays(
 
 
 @st.defines_strategy
-def array_shapes(min_dims=1, max_dims=None, min_side=1, max_side=None):
-    # type: (int, int, int, int) -> st.SearchStrategy[Shape]
+@deprecated_posargs
+def array_shapes(
+    *, min_dims: int = 1, max_dims: int = None, min_side: int = 1, max_side: int = None
+) -> st.SearchStrategy[Shape]:
     """Return a strategy for array shapes (tuples of int >= 1)."""
-    check_type(integer_types, min_dims, "min_dims")
-    check_type(integer_types, min_side, "min_side")
+    check_type(int, min_dims, "min_dims")
+    check_type(int, min_side, "min_side")
     if min_dims > 32:
         raise InvalidArgument(
             "Got min_dims=%r, but numpy does not support arrays greater than 32 dimensions"
@@ -438,7 +427,7 @@ def array_shapes(min_dims=1, max_dims=None, min_side=1, max_side=None):
         )
     if max_dims is None:
         max_dims = min(min_dims + 2, 32)
-    check_type(integer_types, max_dims, "max_dims")
+    check_type(int, max_dims, "max_dims")
     if max_dims > 32:
         raise InvalidArgument(
             "Got max_dims=%r, but numpy does not support arrays greater than 32 dimensions"
@@ -446,7 +435,7 @@ def array_shapes(min_dims=1, max_dims=None, min_side=1, max_side=None):
         )
     if max_side is None:
         max_side = min_side + 5
-    check_type(integer_types, max_side, "max_side")
+    check_type(int, max_side, "max_side")
     order_check("dims", 0, min_dims, max_dims)
     order_check("side", 0, min_side, max_side)
 
@@ -456,8 +445,7 @@ def array_shapes(min_dims=1, max_dims=None, min_side=1, max_side=None):
 
 
 @st.defines_strategy
-def scalar_dtypes():
-    # type: () -> st.SearchStrategy[np.dtype]
+def scalar_dtypes() -> st.SearchStrategy[np.dtype]:
     """Return a strategy that can return any non-flexible scalar dtype."""
     return st.one_of(
         boolean_dtypes(),
@@ -470,8 +458,7 @@ def scalar_dtypes():
     )
 
 
-def defines_dtype_strategy(strat):
-    # type: (T) -> T
+def defines_dtype_strategy(strat: T) -> T:
     @st.defines_strategy
     @proxies(strat)
     def inner(*args, **kwargs):
@@ -481,8 +468,7 @@ def defines_dtype_strategy(strat):
 
 
 @defines_dtype_strategy
-def boolean_dtypes():
-    # type: () -> st.SearchStrategy[np.dtype]
+def boolean_dtypes() -> st.SearchStrategy[np.dtype]:
     return st.just("?")
 
 
@@ -491,7 +477,7 @@ def dtype_factory(kind, sizes, valid_sizes, endianness):
     valid_endian = ("?", "<", "=", ">")
     check_argument(
         endianness in valid_endian,
-        u"Unknown endianness: was {}, must be in {}",
+        "Unknown endianness: was {}, must be in {}",
         endianness,
         valid_endian,
     )
@@ -501,7 +487,7 @@ def dtype_factory(kind, sizes, valid_sizes, endianness):
         check_argument(sizes, "Dtype must have at least one possible size.")
         check_argument(
             all(s in valid_sizes for s in sizes),
-            u"Invalid sizes: was {} must be an item or sequence " u"in {}",
+            "Invalid sizes: was {} must be an item or sequence in {}",
             sizes,
             valid_sizes,
         )
@@ -516,8 +502,10 @@ def dtype_factory(kind, sizes, valid_sizes, endianness):
 
 
 @defines_dtype_strategy
-def unsigned_integer_dtypes(endianness="?", sizes=(8, 16, 32, 64)):
-    # type: (str, Sequence[int]) -> st.SearchStrategy[np.dtype]
+@deprecated_posargs
+def unsigned_integer_dtypes(
+    *, endianness: str = "?", sizes: Sequence[int] = (8, 16, 32, 64)
+) -> st.SearchStrategy[np.dtype]:
     """Return a strategy for unsigned integer dtypes.
 
     endianness may be ``<`` for little-endian, ``>`` for big-endian,
@@ -531,8 +519,10 @@ def unsigned_integer_dtypes(endianness="?", sizes=(8, 16, 32, 64)):
 
 
 @defines_dtype_strategy
-def integer_dtypes(endianness="?", sizes=(8, 16, 32, 64)):
-    # type: (str, Sequence[int]) -> st.SearchStrategy[np.dtype]
+@deprecated_posargs
+def integer_dtypes(
+    *, endianness: str = "?", sizes: Sequence[int] = (8, 16, 32, 64)
+) -> st.SearchStrategy[np.dtype]:
     """Return a strategy for signed integer dtypes.
 
     endianness and sizes are treated as for
@@ -542,8 +532,10 @@ def integer_dtypes(endianness="?", sizes=(8, 16, 32, 64)):
 
 
 @defines_dtype_strategy
-def floating_dtypes(endianness="?", sizes=(16, 32, 64)):
-    # type: (str, Sequence[int]) -> st.SearchStrategy[np.dtype]
+@deprecated_posargs
+def floating_dtypes(
+    *, endianness: str = "?", sizes: Sequence[int] = (16, 32, 64)
+) -> st.SearchStrategy[np.dtype]:
     """Return a strategy for floating-point dtypes.
 
     sizes is the size in bits of floating-point number.  Some machines support
@@ -557,8 +549,10 @@ def floating_dtypes(endianness="?", sizes=(16, 32, 64)):
 
 
 @defines_dtype_strategy
-def complex_number_dtypes(endianness="?", sizes=(64, 128)):
-    # type: (str, Sequence[int]) -> st.SearchStrategy[np.dtype]
+@deprecated_posargs
+def complex_number_dtypes(
+    *, endianness: str = "?", sizes: Sequence[int] = (64, 128)
+) -> st.SearchStrategy[np.dtype]:
     """Return a strategy for complex-number dtypes.
 
     sizes is the total size in bits of a complex number, which consists
@@ -572,13 +566,13 @@ def complex_number_dtypes(endianness="?", sizes=(64, 128)):
 def validate_time_slice(max_period, min_period):
     check_argument(
         max_period in TIME_RESOLUTIONS,
-        u"max_period {} must be a valid resolution in {}",
+        "max_period {} must be a valid resolution in {}",
         max_period,
         TIME_RESOLUTIONS,
     )
     check_argument(
         min_period in TIME_RESOLUTIONS,
-        u"min_period {} must be a valid resolution in {}",
+        "min_period {} must be a valid resolution in {}",
         min_period,
         TIME_RESOLUTIONS,
     )
@@ -586,7 +580,7 @@ def validate_time_slice(max_period, min_period):
     end = TIME_RESOLUTIONS.index(min_period) + 1
     check_argument(
         start < end,
-        u"max_period {} must be earlier in sequence {} than " u"min_period {}",
+        "max_period {} must be earlier in sequence {} than min_period {}",
         max_period,
         TIME_RESOLUTIONS,
         min_period,
@@ -595,8 +589,10 @@ def validate_time_slice(max_period, min_period):
 
 
 @defines_dtype_strategy
-def datetime64_dtypes(max_period="Y", min_period="ns", endianness="?"):
-    # type: (str, str, str) -> st.SearchStrategy[np.dtype]
+@deprecated_posargs
+def datetime64_dtypes(
+    *, max_period: str = "Y", min_period: str = "ns", endianness: str = "?"
+) -> st.SearchStrategy[np.dtype]:
     """Return a strategy for datetime64 dtypes, with various precisions from
     year to attosecond."""
     return dtype_factory(
@@ -608,8 +604,10 @@ def datetime64_dtypes(max_period="Y", min_period="ns", endianness="?"):
 
 
 @defines_dtype_strategy
-def timedelta64_dtypes(max_period="Y", min_period="ns", endianness="?"):
-    # type: (str, str, str) -> st.SearchStrategy[np.dtype]
+@deprecated_posargs
+def timedelta64_dtypes(
+    *, max_period: str = "Y", min_period: str = "ns", endianness: str = "?"
+) -> st.SearchStrategy[np.dtype]:
     """Return a strategy for timedelta64 dtypes, with various precisions from
     year to attosecond."""
     return dtype_factory(
@@ -621,8 +619,10 @@ def timedelta64_dtypes(max_period="Y", min_period="ns", endianness="?"):
 
 
 @defines_dtype_strategy
-def byte_string_dtypes(endianness="?", min_len=1, max_len=16):
-    # type: (str, int, int) -> st.SearchStrategy[np.dtype]
+@deprecated_posargs
+def byte_string_dtypes(
+    *, endianness: str = "?", min_len: int = 1, max_len: int = 16
+) -> st.SearchStrategy[np.dtype]:
     """Return a strategy for generating bytestring dtypes, of various lengths
     and byteorder.
 
@@ -630,28 +630,15 @@ def byte_string_dtypes(endianness="?", min_len=1, max_len=16):
     dtypes with length 0 indicate that size is still to be determined, so
     the minimum length for string dtypes is 1.
     """
-    if min_len == 0:
-        note_deprecation(
-            "generating byte string dtypes for unspecified length ('S0') "
-            "is deprecated. min_len will be 1 instead.",
-            since="2019-09-09",
-        )
-        min_len = 1
-    if max_len == 0:
-        note_deprecation(
-            "generating byte string dtypes for unspecified length ('S0') "
-            "is deprecated. max_len will be 1 instead.",
-            since="2019-09-09",
-        )
-        max_len = 1
-
     order_check("len", 1, min_len, max_len)
     return dtype_factory("S", list(range(min_len, max_len + 1)), None, endianness)
 
 
 @defines_dtype_strategy
-def unicode_string_dtypes(endianness="?", min_len=1, max_len=16):
-    # type: (str, int, int) -> st.SearchStrategy[np.dtype]
+@deprecated_posargs
+def unicode_string_dtypes(
+    *, endianness: str = "?", min_len: int = 1, max_len: int = 16
+) -> st.SearchStrategy[np.dtype]:
     """Return a strategy for generating unicode string dtypes, of various
     lengths and byteorder.
 
@@ -659,61 +646,67 @@ def unicode_string_dtypes(endianness="?", min_len=1, max_len=16):
     dtypes with length 0 indicate that size is still to be determined, so
     the minimum length for string dtypes is 1.
     """
-    if min_len == 0:
-        note_deprecation(
-            "generating unicode string dtypes for unspecified length ('U0') "
-            "is deprecated. min_len will be 1 instead.",
-            since="2019-09-09",
-        )
-        min_len = 1
-    if max_len == 0:
-        note_deprecation(
-            "generating unicode string dtypes for unspecified length ('U0') "
-            "is deprecated. max_len will be 1 instead.",
-            since="2019-09-09",
-        )
-        max_len = 1
-
     order_check("len", 1, min_len, max_len)
     return dtype_factory("U", list(range(min_len, max_len + 1)), None, endianness)
 
 
+def _no_title_is_name_of_a_titled_field(ls):
+    seen = set()
+    for title_and_name, *_ in ls:
+        if isinstance(title_and_name, tuple):
+            if seen.intersection(title_and_name):  # pragma: no cover
+                # Our per-element filters below make this as rare as possible,
+                # so it's not always covered.
+                return False
+            seen.update(title_and_name)
+    return True
+
+
 @defines_dtype_strategy
+@deprecated_posargs
 def array_dtypes(
-    subtype_strategy=scalar_dtypes(),  # type: st.SearchStrategy[np.dtype]
-    min_size=1,  # type: int
-    max_size=5,  # type: int
-    allow_subarrays=False,  # type: bool
-):
-    # type: (...) -> st.SearchStrategy[np.dtype]
+    subtype_strategy: st.SearchStrategy[np.dtype] = scalar_dtypes(),
+    *,
+    min_size: int = 1,
+    max_size: int = 5,
+    allow_subarrays: bool = False
+) -> st.SearchStrategy[np.dtype]:
     """Return a strategy for generating array (compound) dtypes, with members
     drawn from the given subtype strategy."""
     order_check("size", 0, min_size, max_size)
-    # Field names must be native strings and the empty string is weird; see #1963.
-    if PY2:
-        field_names = st.binary(min_size=1)
-    else:
-        field_names = st.text(min_size=1)
-    elements = st.tuples(field_names, subtype_strategy)
+    # The empty string is replaced by f{idx}; see #1963 for details.  Much easier to
+    # insist that field names be unique and just boost f{idx} strings manually.
+    field_names = st.integers(0, 127).map("f{}".format) | st.text(min_size=1)
+    name_titles = st.one_of(
+        field_names,
+        st.tuples(field_names, field_names).filter(lambda ns: ns[0] != ns[1]),
+    )
+    elements = st.tuples(name_titles, subtype_strategy)
     if allow_subarrays:
         elements |= st.tuples(
-            field_names, subtype_strategy, array_shapes(max_dims=2, max_side=2)
+            name_titles, subtype_strategy, array_shapes(max_dims=2, max_side=2)
         )
     return st.lists(
         elements=elements,
         min_size=min_size,
         max_size=max_size,
-        unique_by=lambda d: d[0],
-    )
+        unique_by=(
+            # Deduplicate by both name and title for efficiency before filtering.
+            # (Field names must be unique, as must titles, and no intersections)
+            lambda d: d[0] if isinstance(d[0], str) else d[0][0],
+            lambda d: d[0] if isinstance(d[0], str) else d[0][1],
+        ),
+    ).filter(_no_title_is_name_of_a_titled_field)
 
 
 @st.defines_strategy
+@deprecated_posargs
 def nested_dtypes(
-    subtype_strategy=scalar_dtypes(),  # type: st.SearchStrategy[np.dtype]
-    max_leaves=10,  # type: int
-    max_itemsize=None,  # type: int
-):
-    # type: (...) -> st.SearchStrategy[np.dtype]
+    subtype_strategy: st.SearchStrategy[np.dtype] = scalar_dtypes(),
+    *,
+    max_leaves: int = 10,
+    max_itemsize: int = None
+) -> st.SearchStrategy[np.dtype]:
     """Return the most-general dtype strategy.
 
     Elements drawn from this strategy may be simple (from the
@@ -723,13 +716,17 @@ def nested_dtypes(
     argument.
     """
     return st.recursive(
-        subtype_strategy, lambda x: array_dtypes(x, allow_subarrays=True), max_leaves
+        subtype_strategy,
+        lambda x: array_dtypes(x, allow_subarrays=True),
+        max_leaves=max_leaves,
     ).filter(lambda d: max_itemsize is None or d.itemsize <= max_itemsize)
 
 
 @st.defines_strategy
-def valid_tuple_axes(ndim, min_size=0, max_size=None):
-    # type: (int, int, int) -> st.SearchStrategy[Shape]
+@deprecated_posargs
+def valid_tuple_axes(
+    ndim: int, *, min_size: int = 0, max_size: int = None
+) -> st.SearchStrategy[Shape]:
     """Return a strategy for generating permissible tuple-values for the
     ``axis`` argument for a numpy sequential function (e.g.
     :func:`numpy:numpy.sum`), given an array of the specified
@@ -759,9 +756,9 @@ def valid_tuple_axes(ndim, min_size=0, max_size=None):
     if max_size is None:
         max_size = ndim
 
-    check_type(integer_types, ndim, "ndim")
-    check_type(integer_types, min_size, "min_size")
-    check_type(integer_types, max_size, "max_size")
+    check_type(int, ndim, "ndim")
+    check_type(int, min_size, "min_size")
+    check_type(int, max_size, "max_size")
     order_check("size", 0, min_size, max_size)
     check_valid_interval(max_size, ndim, "max_size", "ndim")
 
@@ -769,12 +766,21 @@ def valid_tuple_axes(ndim, min_size=0, max_size=None):
     axes = st.integers(0, max(0, 2 * ndim - 1)).map(
         lambda x: x if x < ndim else x - 2 * ndim
     )
-    return st.lists(axes, min_size, max_size, unique_by=lambda x: x % ndim).map(tuple)
+    return st.lists(
+        axes, min_size=min_size, max_size=max_size, unique_by=lambda x: x % ndim
+    ).map(tuple)
 
 
 @st.defines_strategy
-def broadcastable_shapes(shape, min_dims=0, max_dims=None, min_side=1, max_side=None):
-    # type: (Shape, int, int, int, int) -> st.SearchStrategy[Shape]
+@deprecated_posargs
+def broadcastable_shapes(
+    shape: Shape,
+    *,
+    min_dims: int = 0,
+    max_dims: int = None,
+    min_side: int = 1,
+    max_side: int = None
+) -> st.SearchStrategy[Shape]:
     """Return a strategy for generating shapes that are broadcast-compatible
     with the provided shape.
 
@@ -800,18 +806,18 @@ def broadcastable_shapes(shape, min_dims=0, max_dims=None, min_side=1, max_side=
     """
     check_type(tuple, shape, "shape")
     strict_check = max_side is None or max_dims is None
-    check_type(integer_types, min_side, "min_side")
-    check_type(integer_types, min_dims, "min_dims")
+    check_type(int, min_side, "min_side")
+    check_type(int, min_dims, "min_dims")
 
     if max_dims is None:
         max_dims = min(32, max(len(shape), min_dims) + 2)
     else:
-        check_type(integer_types, max_dims, "max_dims")
+        check_type(int, max_dims, "max_dims")
 
     if max_side is None:
         max_side = max(tuple(shape[-max_dims:]) + (min_side,)) + 2
     else:
-        check_type(integer_types, max_side, "max_side")
+        check_type(int, max_side, "max_side")
 
     order_check("dims", 0, min_dims, max_dims)
     order_check("side", 0, min_side, max_side)
@@ -1026,7 +1032,9 @@ _ARGUMENT_LIST = "{0}(?:,{0})*".format(_SHAPE)
 _SIGNATURE = r"^{}->{}$".format(_ARGUMENT_LIST, _SHAPE)
 _SIGNATURE_MULTIPLE_OUTPUT = r"^{0}->{0}$".format(_ARGUMENT_LIST)
 
-_GUfuncSig = namedtuple("_GUfuncSig", ["input_shapes", "result_shape"])
+_GUfuncSig = NamedTuple(
+    "_GUfuncSig", [("input_shapes", Tuple[Shape, ...]), ("result_shape", Shape)]
+)
 
 
 def _hypothesis_parse_gufunc_signature(signature, all_checks=True):
@@ -1077,23 +1085,23 @@ def _hypothesis_parse_gufunc_signature(signature, all_checks=True):
                         raise InvalidArgument(frozen_optional_err % (name, signature))
                 except ValueError:
                     if name.strip("?") in (names_out - names_in):
-                        quiet_raise(InvalidArgument(only_out_err % (name, signature)))
+                        raise InvalidArgument(
+                            only_out_err % (name, signature)
+                        ) from None
     return _GUfuncSig(input_shapes=input_shapes, result_shape=result_shape)
 
 
 @st.defines_strategy
-@reserved_means_kwonly_star
 def mutually_broadcastable_shapes(
-    __reserved=not_set,  # type: Any
-    num_shapes=not_set,  # type: Union[UniqueIdentifier, int]
-    signature=not_set,  # type: Union[UniqueIdentifier, str]
-    base_shape=(),  # type: Shape
-    min_dims=0,  # type: int
-    max_dims=None,  # type: int
-    min_side=1,  # type: int
-    max_side=None,  # type: int
-):
-    # type: (...) -> st.SearchStrategy[BroadcastableShapes]
+    *,
+    num_shapes: Union[UniqueIdentifier, int] = not_set,
+    signature: Union[UniqueIdentifier, str] = not_set,
+    base_shape: Shape = (),
+    min_dims: int = 0,
+    max_dims: int = None,
+    min_side: int = 1,
+    max_side: int = None
+) -> st.SearchStrategy[BroadcastableShapes]:
     """Return a strategy for generating a specified number of shapes, N, that are
     mutually-broadcastable with one another and with the provided "base-shape".
 
@@ -1167,14 +1175,11 @@ def mutually_broadcastable_shapes(
         BroadcastableShapes(input_shapes=((3, 4, 2), (1, 2)), result_shape=(3, 4))
         BroadcastableShapes(input_shapes=((4, 2), (1, 2, 3)), result_shape=(4, 3))
     """
-    if __reserved is not not_set:
-        raise InvalidArgument("Do not pass the __reserved argument.")
-
     arg_msg = "Pass either the `num_shapes` or the `signature` argument, but not both."
     if num_shapes is not not_set:
         check_argument(signature is not_set, arg_msg)
-        check_type(integer_types, num_shapes, "num_shapes")
-        assert isinstance(num_shapes, integer_types)  # for mypy
+        check_type(int, num_shapes, "num_shapes")
+        assert isinstance(num_shapes, int)  # for mypy
         check_argument(num_shapes >= 1, "num_shapes={} must be at least 1", num_shapes)
         parsed_signature = None
         sig_dims = 0
@@ -1185,7 +1190,7 @@ def mutually_broadcastable_shapes(
                 "Expected a string, but got invalid signature=None.  "
                 "(maybe .signature attribute of an element-wise ufunc?)"
             )
-        check_type(string_types, signature, "signature")
+        check_type(str, signature, "signature")
         parsed_signature = _hypothesis_parse_gufunc_signature(signature)
         sig_dims = min(
             map(len, parsed_signature.input_shapes + (parsed_signature.result_shape,))
@@ -1195,18 +1200,18 @@ def mutually_broadcastable_shapes(
 
     check_type(tuple, base_shape, "base_shape")
     strict_check = max_dims is not None
-    check_type(integer_types, min_side, "min_side")
-    check_type(integer_types, min_dims, "min_dims")
+    check_type(int, min_side, "min_side")
+    check_type(int, min_dims, "min_dims")
 
     if max_dims is None:
         max_dims = min(32 - sig_dims, max(len(base_shape), min_dims) + 2)
     else:
-        check_type(integer_types, max_dims, "max_dims")
+        check_type(int, max_dims, "max_dims")
 
     if max_side is None:
         max_side = max(tuple(base_shape[-max_dims:]) + (min_side,)) + 2
     else:
-        check_type(integer_types, max_side, "max_side")
+        check_type(int, max_side, "max_side")
 
     order_check("dims", 0, min_dims, max_dims)
     order_check("side", 0, min_side, max_side)
@@ -1310,16 +1315,14 @@ class BasicIndexStrategy(SearchStrategy):
 
 
 @st.defines_strategy
-@reserved_means_kwonly_star
 def basic_indices(
-    shape,
-    __reserved=not_set,
-    min_dims=0,
-    max_dims=None,
-    allow_newaxis=False,
-    allow_ellipsis=True,
-):
-    # type: (Shape, Any, int, int, bool, bool) -> st.SearchStrategy[BasicIndex]
+    shape: Shape,
+    *,
+    min_dims: int = 0,
+    max_dims: int = None,
+    allow_newaxis: bool = False,
+    allow_ellipsis: bool = True
+) -> st.SearchStrategy[BasicIndex]:
     """
     The ``basic_indices`` strategy generates `basic indexes
     <https://docs.scipy.org/doc/numpy/reference/arrays.indexing.html>`__  for
@@ -1347,24 +1350,22 @@ def basic_indices(
     # Arguments to exclude scalars, zero-dim arrays, and dims of size zero were
     # all considered and rejected.  We want users to explicitly consider those
     # cases if they're dealing in general indexers, and while it's fiddly we can
-    # back-compatibly add them later (hence using __reserved to sim kwonlyargs).
+    # back-compatibly add them later (hence using kwonlyargs).
     check_type(tuple, shape, "shape")
-    if __reserved is not not_set:
-        raise InvalidArgument("Do not pass the __reserved argument.")
     check_type(bool, allow_ellipsis, "allow_ellipsis")
     check_type(bool, allow_newaxis, "allow_newaxis")
-    check_type(integer_types, min_dims, "min_dims")
+    check_type(int, min_dims, "min_dims")
     if max_dims is None:
         max_dims = min(max(len(shape), min_dims) + 2, 32)
     else:
-        check_type(integer_types, max_dims, "max_dims")
+        check_type(int, max_dims, "max_dims")
     order_check("dims", 0, min_dims, max_dims)
     check_argument(
         max_dims <= 32,
         "max_dims=%r, but numpy arrays have at most 32 dimensions" % (max_dims,),
     )
     check_argument(
-        all(isinstance(x, integer_types) and x >= 0 for x in shape),
+        all(isinstance(x, int) and x >= 0 for x in shape),
         "shape=%r, but all dimensions must be of integer size >= 0" % (shape,),
     )
     return BasicIndexStrategy(
@@ -1377,8 +1378,13 @@ def basic_indices(
 
 
 @st.defines_strategy
-def integer_array_indices(shape, result_shape=array_shapes(), dtype="int"):
-    # type: (Shape, SearchStrategy[Shape], np.dtype) -> st.SearchStrategy[Tuple[np.ndarray, ...]]
+@deprecated_posargs
+def integer_array_indices(
+    shape: Shape,
+    *,
+    result_shape: SearchStrategy[Shape] = array_shapes(),
+    dtype: np.dtype = "int"
+) -> st.SearchStrategy[Tuple[np.ndarray, ...]]:
     """Return a search strategy for tuples of integer-arrays that, when used
     to index into an array of shape ``shape``, given an array whose shape
     was drawn from ``result_shape``.
@@ -1420,7 +1426,7 @@ def integer_array_indices(shape, result_shape=array_shapes(), dtype="int"):
     """
     check_type(tuple, shape, "shape")
     check_argument(
-        shape and all(isinstance(x, integer_types) and x > 0 for x in shape),
+        shape and all(isinstance(x, int) and x > 0 for x in shape),
         "shape=%r must be a non-empty tuple of integers > 0" % (shape,),
     )
     check_type(SearchStrategy, result_shape, "result_shape")
