@@ -1,9 +1,7 @@
-# coding=utf-8
-#
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis/
 #
-# Most of this work is copyright (C) 2013-2019 David R. MacIver
+# Most of this work is copyright (C) 2013-2020 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
 # CONTRIBUTING.rst for a full list of people who may hold copyright, and
 # consult the git log if you need to determine who owns an individual
@@ -15,8 +13,6 @@
 #
 # END HEADER
 
-from __future__ import absolute_import, division, print_function
-
 from collections import Counter, defaultdict
 from enum import Enum
 from random import Random, getrandbits
@@ -27,7 +23,7 @@ import attr
 from hypothesis import HealthCheck, Phase, Verbosity, settings as Settings
 from hypothesis._settings import local_settings
 from hypothesis.internal.cache import LRUReusedCache
-from hypothesis.internal.compat import ceil, hbytes, int_from_bytes
+from hypothesis.internal.compat import ceil, int_from_bytes
 from hypothesis.internal.conjecture.data import (
     ConjectureData,
     ConjectureResult,
@@ -42,7 +38,7 @@ from hypothesis.internal.conjecture.datatree import (
     TreeRecordingObserver,
 )
 from hypothesis.internal.conjecture.junkdrawer import clamp
-from hypothesis.internal.conjecture.pareto import NO_SCORE, ParetoFront
+from hypothesis.internal.conjecture.pareto import NO_SCORE, ParetoFront, ParetoOptimiser
 from hypothesis.internal.conjecture.shrinker import Shrinker, sort_key
 from hypothesis.internal.healthcheck import fail_health_check
 from hypothesis.reporting import base_report
@@ -60,7 +56,7 @@ BUFFER_SIZE = 8 * 1024
 
 
 @attr.s
-class HealthCheckState(object):
+class HealthCheckState:
     valid_examples = attr.ib(default=0)
     invalid_examples = attr.ib(default=0)
     overrun_examples = attr.ib(default=0)
@@ -79,7 +75,7 @@ class RunIsComplete(Exception):
     pass
 
 
-class ConjectureRunner(object):
+class ConjectureRunner:
     def __init__(self, test_function, settings=None, random=None, database_key=None):
         self._test_function = test_function
         self.settings = settings or Settings()
@@ -345,7 +341,7 @@ class ConjectureRunner(object):
             key = self.sub_key(sub_key)
             if key is None:
                 return
-            self.settings.database.save(key, hbytes(buffer))
+            self.settings.database.save(key, bytes(buffer))
 
     def downgrade_buffer(self, buffer):
         if self.settings.database is not None and self.database_key is not None:
@@ -427,7 +423,7 @@ class ConjectureRunner(object):
             for v in self.interesting_examples.values():
                 self.debug_data(v)
             self.debug(
-                u"Run complete after %d examples (%d valid) and %d shrinks"
+                "Run complete after %d examples (%d valid) and %d shrinks"
                 % (self.call_count, self.valid_examples, self.shrinks)
             )
 
@@ -509,6 +505,34 @@ class ConjectureRunner(object):
         self.exit_reason = reason
         raise RunIsComplete()
 
+    def should_generate_more(self):
+        # End the generation phase where we would have ended it if no bugs had
+        # been found.  This reproduces the exit logic in `self.test_function`,
+        # but with the important distinction that this clause will move on to
+        # the shrinking phase having found one or more bugs, while the other
+        # will exit having found zero bugs.
+        if self.valid_examples >= self.settings.max_examples or self.call_count >= max(
+            self.settings.max_examples * 10, 1000
+        ):  # pragma: no cover
+            return False
+
+        # If we haven't found a bug, keep looking - if we hit any limits on
+        # the number of tests to run that will raise an exception and stop
+        # the run.
+        if not self.interesting_examples:
+            return True
+        # If we've found a bug and won't report more than one, stop looking.
+        elif not self.settings.report_multiple_bugs:
+            return False
+        assert self.first_bug_found_at <= self.last_bug_found_at <= self.call_count
+        # Otherwise, keep searching for between ten and 'a heuristic' calls.
+        # We cap 'calls after first bug' so errors are reported reasonably
+        # soon even for tests that are allowed to run for a very long time,
+        # or sooner if the latest half of our test effort has been fruitless.
+        return self.call_count < MIN_TEST_CALLS or self.call_count < min(
+            self.first_bug_found_at + 1000, self.last_bug_found_at * 2
+        )
+
     def generate_new_examples(self):
         if Phase.generate not in self.settings.phases:
             return
@@ -520,7 +544,8 @@ class ConjectureRunner(object):
 
         self.debug("Generating new examples")
 
-        zero_data = self.cached_test_function(hbytes(BUFFER_SIZE))
+        assert self.should_generate_more()
+        zero_data = self.cached_test_function(bytes(BUFFER_SIZE))
         if zero_data.status > Status.OVERRUN:
             self.__data_cache.pin(zero_data.buffer)
 
@@ -544,40 +569,6 @@ class ConjectureRunner(object):
 
         self.health_check_state = HealthCheckState()
 
-        def should_generate_more():
-            # End the generation phase where we would have ended it if no bugs had
-            # been found.  This reproduces the exit logic in `self.test_function`,
-            # but with the important distinction that this clause will move on to
-            # the shrinking phase having found one or more bugs, while the other
-            # will exit having found zero bugs.
-            if (
-                self.valid_examples >= self.settings.max_examples
-                or self.call_count >= max(self.settings.max_examples * 10, 1000)
-                or (
-                    self.best_examples_of_observed_targets
-                    and self.valid_examples * 2 >= self.settings.max_examples
-                    and self.should_optimise
-                )
-            ):  # pragma: no cover
-                return False
-
-            # If we haven't found a bug, keep looking - if we hit any limits on
-            # the number of tests to run that will raise an exception and stop
-            # the run.
-            if not self.interesting_examples:
-                return True
-            # If we've found a bug and won't report more than one, stop looking.
-            elif not self.settings.report_multiple_bugs:
-                return False
-            assert self.first_bug_found_at <= self.last_bug_found_at <= self.call_count
-            # Otherwise, keep searching for between ten and 'a heuristic' calls.
-            # We cap 'calls after first bug' so errors are reported reasonably
-            # soon even for tests that are allowed to run for a very long time,
-            # or sooner if the latest half of our test effort has been fruitless.
-            return self.call_count < MIN_TEST_CALLS or self.call_count < min(
-                self.first_bug_found_at + 1000, self.last_bug_found_at * 2
-            )
-
         # We attempt to use the size of the minimal generated test case starting
         # from a given novel prefix as a guideline to generate smaller test
         # cases for an initial period, by restriscting ourselves to test cases
@@ -594,30 +585,32 @@ class ConjectureRunner(object):
         # it has failed too many times in a row.
         consecutive_zero_extend_is_invalid = 0
 
-        while should_generate_more():
+        # We control growth during initial example generation, for two
+        # reasons:
+        #
+        # * It gives us an opportunity to find small examples early, which
+        #   gives us a fast path for easy to find bugs.
+        # * It avoids low probability events where we might end up
+        #   generating very large examples during health checks, which
+        #   on slower machines can trigger HealthCheck.too_slow.
+        #
+        # The heuristic we use is that we attempt to estimate the smallest
+        # extension of this prefix, and limit the size to no more than
+        # an order of magnitude larger than that. If we fail to estimate
+        # the size accurately, we skip over this prefix and try again.
+        #
+        # We need to tune the example size based on the initial prefix,
+        # because any fixed size might be too small, and any size based
+        # on the strategy in general can fall afoul of strategies that
+        # have very different sizes for different prefixes.
+        small_example_cap = clamp(10, self.settings.max_examples // 10, 50)
+
+        optimise_at = max(self.settings.max_examples // 2, small_example_cap + 1)
+        ran_optimisations = False
+
+        while self.should_generate_more():
             prefix = self.generate_novel_prefix()
             assert len(prefix) <= BUFFER_SIZE
-
-            # We control growth during initial example generation, for two
-            # reasons:
-            #
-            # * It gives us an opportunity to find small examples early, which
-            #   gives us a fast path for easy to find bugs.
-            # * It avoids low probability events where we might end up
-            #   generating very large examples during health checks, which
-            #   on slower machines can trigger HealthCheck.too_slow.
-            #
-            # The heuristic we use is that we attempt to estimate the smallest
-            # extension of this prefix, and limit the size to no more than
-            # an order of magnitude larger than that. If we fail to estimate
-            # the size accurately, we skip over this prefix and try again.
-            #
-            # We need to tune the example size based on the initial prefix,
-            # because any fixed size might be too small, and any size based
-            # on the strategy in general can fall afoul of strategies that
-            # have very different sizes for different prefixes.
-            small_example_cap = clamp(10, self.settings.max_examples // 10, 50)
-
             if (
                 self.valid_examples <= small_example_cap
                 and self.call_count <= 5 * small_example_cap
@@ -625,7 +618,7 @@ class ConjectureRunner(object):
                 and consecutive_zero_extend_is_invalid < 5
             ):
                 minimal_example = self.cached_test_function(
-                    prefix + hbytes(BUFFER_SIZE - len(prefix))
+                    prefix + bytes(BUFFER_SIZE - len(prefix))
                 )
 
                 if minimal_example.status < Status.VALID:
@@ -662,7 +655,7 @@ class ConjectureRunner(object):
 
                 # We might have hit the cap on number of examples we should
                 # run when calculating the minimal example.
-                if not should_generate_more():
+                if not self.should_generate_more():
                     break
 
                 prefix = trial_data.buffer
@@ -673,112 +666,158 @@ class ConjectureRunner(object):
 
             self.test_function(data)
 
-            # A thing that is often useful but rarely happens by accident is
-            # to generate the same value at multiple different points in the
-            # test case.
-            #
-            # Rather than make this the responsibility of individual strategies
-            # we implement a small mutator that just takes parts of the test
-            # case with the same label and tries replacing one of them with a
-            # copy of the other and tries running it. If we've made a good
-            # guess about what to put where, this will run a similar generated
-            # test case with more duplication.
+            self.generate_mutations_from(data)
+
+            # Although the optimisations are logically a distinct phase, we
+            # actually normally run them as part of example generation. The
+            # reason for this is that we cannot guarantee that optimisation
+            # actually exhausts our budget: It might finish running and we
+            # discover that actually we still could run a bunch more test cases
+            # if we want.
             if (
-                # An OVERRUN doesn't have enough information about the test
-                # case to mutate, so we just skip those.
-                data.status >= Status.INVALID
-                # This has a tendency to trigger some weird edge cases during
-                # generation so we don't let it run until we're done with the
-                # health checks.
-                and self.health_check_state is None
+                self.valid_examples >= max(small_example_cap, optimise_at)
+                and not ran_optimisations
             ):
-                initial_calls = self.call_count
-                failed_mutations = 0
-                while (
-                    should_generate_more()
-                    # We implement fairly conservative checks for how long we
-                    # we should run mutation for, as it's generally not obvious
-                    # how helpful it is for any given test case.
-                    and self.call_count <= initial_calls + 5
-                    and failed_mutations <= 5
-                ):
-                    groups = defaultdict(list)
-                    for ex in data.examples:
-                        groups[ex.label, ex.depth].append(ex)
+                ran_optimisations = True
+                self.optimise_targets()
 
-                    groups = [v for v in groups.values() if len(v) > 1]
+    def generate_mutations_from(self, data):
+        # A thing that is often useful but rarely happens by accident is
+        # to generate the same value at multiple different points in the
+        # test case.
+        #
+        # Rather than make this the responsibility of individual strategies
+        # we implement a small mutator that just takes parts of the test
+        # case with the same label and tries replacing one of them with a
+        # copy of the other and tries running it. If we've made a good
+        # guess about what to put where, this will run a similar generated
+        # test case with more duplication.
+        if (
+            # An OVERRUN doesn't have enough information about the test
+            # case to mutate, so we just skip those.
+            data.status >= Status.INVALID
+            # This has a tendency to trigger some weird edge cases during
+            # generation so we don't let it run until we're done with the
+            # health checks.
+            and self.health_check_state is None
+        ):
+            initial_calls = self.call_count
+            failed_mutations = 0
 
-                    if not groups:
-                        break
+            while (
+                self.should_generate_more()
+                # We implement fairly conservative checks for how long we
+                # we should run mutation for, as it's generally not obvious
+                # how helpful it is for any given test case.
+                and self.call_count <= initial_calls + 5
+                and failed_mutations <= 5
+            ):
+                groups = data.examples.mutator_groups
+                if not groups:
+                    break
 
-                    group = self.random.choice(groups)
+                group = self.random.choice(groups)
 
-                    ex1, ex2 = sorted(
-                        self.random.sample(group, 2), key=lambda i: i.index
+                ex1, ex2 = [
+                    data.examples[i] for i in sorted(self.random.sample(group, 2))
+                ]
+                assert ex1.end <= ex2.start
+
+                replacements = [data.buffer[e.start : e.end] for e in [ex1, ex2]]
+
+                replacement = self.random.choice(replacements)
+
+                try:
+                    # We attempt to replace both the the examples with
+                    # whichever choice we made. Note that this might end
+                    # up messing up and getting the example boundaries
+                    # wrong - labels matching are only a best guess as to
+                    # whether the two are equivalent - but it doesn't
+                    # really matter. It may not achieve the desired result
+                    # but it's still a perfectly acceptable choice sequence.
+                    # to try.
+                    new_data = self.cached_test_function(
+                        data.buffer[: ex1.start]
+                        + replacement
+                        + data.buffer[ex1.end : ex2.start]
+                        + replacement
+                        + data.buffer[ex2.end :],
+                        # We set error_on_discard so that we don't end up
+                        # entering parts of the tree we consider redundant
+                        # and not worth exploring.
+                        error_on_discard=True,
+                        extend=BUFFER_SIZE,
                     )
-                    assert ex1.end <= ex2.start
+                except ContainsDiscard:
+                    failed_mutations += 1
+                    continue
 
-                    replacements = [data.buffer[e.start : e.end] for e in [ex1, ex2]]
-
-                    replacement = self.random.choice(replacements)
-
-                    try:
-                        # We attempt to replace both the the examples with
-                        # whichever choice we made. Note that this might end
-                        # up messing up and getting the example boundaries
-                        # wrong - labels matching are only a best guess as to
-                        # whether the two are equivalent - but it doesn't
-                        # really matter. It may not achieve the desired result
-                        # but it's still a perfectly acceptable choice sequence.
-                        # to try.
-                        new_data = self.cached_test_function(
-                            data.buffer[: ex1.start]
-                            + replacement
-                            + data.buffer[ex1.end : ex2.start]
-                            + replacement
-                            + data.buffer[ex2.end :],
-                            # We set error_on_discard so that we don't end up
-                            # entering parts of the tree we consider redundant
-                            # and not worth exploring.
-                            error_on_discard=True,
-                            extend=BUFFER_SIZE,
-                        )
-                    except ContainsDiscard:
-                        failed_mutations += 1
-                        continue
-
-                    if (
-                        new_data.status >= data.status
-                        and data.buffer != new_data.buffer
-                        and all(
-                            k in new_data.target_observations
-                            and new_data.target_observations[k] >= v
-                            for k, v in data.target_observations.items()
-                        )
-                    ):
-                        data = new_data
-                        failed_mutations = 0
-                    else:
-                        failed_mutations += 1
+                if (
+                    new_data.status >= data.status
+                    and data.buffer != new_data.buffer
+                    and all(
+                        k in new_data.target_observations
+                        and new_data.target_observations[k] >= v
+                        for k, v in data.target_observations.items()
+                    )
+                ):
+                    data = new_data
+                    failed_mutations = 0
+                else:
+                    failed_mutations += 1
 
     def optimise_targets(self):
         """If any target observations have been made, attempt to optimise them
         all."""
         if not self.should_optimise:
             return
+        from hypothesis.internal.conjecture.optimiser import Optimiser
+
+        # We want to avoid running the optimiser for too long in case we hit
+        # an unbounded target score. We start this off fairly conservatively
+        # in case interesting examples are easy to find and then ramp it up
+        # on an exponential schedule so we don't hamper the optimiser too much
+        # if it needs a long time to find good enough improvements.
+        max_improvements = 10
         while True:
             prev_calls = self.call_count
+
+            any_improvements = False
+
             for target, data in list(self.best_examples_of_observed_targets.items()):
-                self.new_optimiser(data, target).run()
-            # This shows up as a missing branch for some reason but is definitely
-            # covered if you add an assertion to check.
-            if self.interesting_examples or (prev_calls == self.call_count):
+                optimiser = Optimiser(
+                    self, data, target, max_improvements=max_improvements
+                )
+                optimiser.run()
+                if optimiser.improvements > 0:
+                    any_improvements = True
+
+            if self.interesting_examples:
                 break
 
+            max_improvements *= 2
+
+            if any_improvements:
+                continue
+
+            self.pareto_optimise()
+
+            if prev_calls == self.call_count:
+                break
+
+    def pareto_optimise(self):
+        if self.pareto_front is not None:
+            ParetoOptimiser(self).run()
+
     def _run(self):
         self.reuse_existing_examples()
         self.generate_new_examples()
-        self.optimise_targets()
+
+        # We normally run the targeting phase mixed in with the generate phase,
+        # but if we've been asked to run it but not generation then we have to
+        # run it explciitly on its own here.
+        if Phase.generate not in self.settings.phases:
+            self.optimise_targets()
         self.shrink_interesting_examples()
         self.exit_with(ExitReason.finished)
 
@@ -868,18 +907,13 @@ class ConjectureRunner(object):
                     # of this reason for interestingness.
                     self.settings.database.delete(self.secondary_key, c)
 
-    def shrink(self, example, predicate):
-        s = self.new_shrinker(example, predicate)
+    def shrink(self, example, predicate=None, allow_transition=None):
+        s = self.new_shrinker(example, predicate, allow_transition)
         s.shrink()
         return s.shrink_target
 
-    def new_shrinker(self, example, predicate):
-        return Shrinker(self, example, predicate)
-
-    def new_optimiser(self, example, target):
-        from hypothesis.internal.conjecture.optimiser import Optimiser
-
-        return Optimiser(self, example, target)
+    def new_shrinker(self, example, predicate=None, allow_transition=None):
+        return Shrinker(self, example, predicate, allow_transition)
 
     def cached_test_function(self, buffer, error_on_discard=False, extend=0):
         """Checks the tree to see if we've tested this buffer, and returns the
@@ -895,7 +929,7 @@ class ConjectureRunner(object):
         result has discards if we cannot determine from previous runs whether
         it will have a discard.
         """
-        buffer = hbytes(buffer)[:BUFFER_SIZE]
+        buffer = bytes(buffer)[:BUFFER_SIZE]
 
         max_length = min(BUFFER_SIZE, len(buffer) + extend)
 
